
"""Evolution programming in python with talib and gplearn inspired API
``tagenalgo`` is a parameter optimization algorithm for trading strategy.
"""

from __future__ import absolute_import
from .tagenalgo import *
from .backtest import *
from .utils import *

__version__ = '1.0.3'
__license__ = 'BSD'